"""
Topology class for optical network topologies.

This module defines the Topology class, representing a network topology with nodes and links,
and providing an adjacency matrix using numpy.

This file uses NetworkX (BSD 3-Clause License):
https://github.com/networkx/networkx/blob/main/LICENSE.txt
"""

from typing import List, Dict, Any
import numpy as np
import networkx as nx
from topolib.elements.node import Node
from topolib.elements.link import Link


class Topology:
    """
    Represents a network topology with nodes and links.

    :param nodes: Initial list of nodes (optional).
    :type nodes: list[topolib.elements.node.Node] or None
    :param links: Initial list of links (optional).
    :type links: list[topolib.elements.link.Link] or None

    :ivar nodes: List of nodes in the topology.
    :vartype nodes: list[Node]
    :ivar links: List of links in the topology.
    :vartype links: list[Link]

    **Examples**
        >>> from topolib.elements.node import Node
        >>> from topolib.elements.link import Link
        >>> from topolib.topology import Topology
        >>> n1 = Node(1, "A", 0.0, 0.0)
        >>> n2 = Node(2, "B", 1.0, 1.0)
        >>> l1 = Link(1, n1, n2, 10.0)
        >>> topo = Topology(nodes=[n1, n2], links=[l1])
        >>> topo.adjacency_matrix()
        array([[0, 1],
               [1, 0]])
    """

    def __init__(self, nodes: List[Node] = None, links: List[Link] = None):
        """
        Initialize a Topology object.

    :param nodes: Initial list of nodes (optional).
    :type nodes: list[topolib.elements.node.Node] or None
    :param links: Initial list of links (optional).
    :type links: list[topolib.elements.link.Link] or None
        """
        self.nodes: List[Node] = nodes if nodes is not None else []
        self.links: List[Link] = links if links is not None else []
        # Internal NetworkX graph for algorithms and visualization
        self._graph = nx.Graph()
        for node in self.nodes:
            self._graph.add_node(node.id, node=node)
        for link in self.links:
            self._graph.add_edge(link.source.id, link.target.id, link=link)

    def add_node(self, node: Node) -> None:
        """
        Add a node to the topology.

        :param node: Node to add.
        :type node: Node
        """
        self.nodes.append(node)
        self._graph.add_node(node.id, node=node)

    def add_link(self, link: Link) -> None:
        """
        Add a link to the topology.

        :param link: Link to add.
        :type link: Link
        """
        self.links.append(link)
        self._graph.add_edge(link.source.id, link.target.id, link=link)

    def remove_node(self, node_id: int) -> None:
        """
        Remove a node and all its links by node id.

        :param node_id: ID of the node to remove.
        :type node_id: int
        """
        self.nodes = [n for n in self.nodes if n.id != node_id]
        self.links = [
            l for l in self.links if l.source.id != node_id and l.target.id != node_id
        ]
        self._graph.remove_node(node_id)

    def remove_link(self, link_id: int) -> None:
        """
        Remove a link by its id.

        :param link_id: ID of the link to remove.
        :type link_id: int
        """
        # Find the link and remove from graph
        link = next((l for l in self.links if l.id == link_id), None)
        if link:
            self._graph.remove_edge(link.source.id, link.target.id)
        self.links = [l for l in self.links if l.id != link_id]

    def adjacency_matrix(self) -> np.ndarray:
        """
        Return the adjacency matrix of the topology as a numpy array.

        :return: Adjacency matrix (1 if connected, 0 otherwise).
        :rtype: numpy.ndarray

        **Example**
            >>> topo.adjacency_matrix()
            array([[0, 1],
                   [1, 0]])
        """
        # Usa NetworkX para obtener la matriz de adyacencia
        if not self.nodes:
            return np.zeros((0, 0), dtype=int)
        node_ids = [n.id for n in self.nodes]
        mat = nx.to_numpy_array(self._graph, nodelist=node_ids, dtype=int)
        return mat



"""
Metrics module for network topology analysis.
"""

from typing import List, Any, Dict, Optional


from topolib.topology import Topology


class Metrics:
    """
    Provides static methods for computing metrics on network topologies.

    Todos los métodos reciben una instancia de Topology.

    Métodos
    -------
    node_degree(topology)
        Calcula el grado de cada nodo.
    link_length_stats(topology)
        Calcula estadísticas (min, max, avg) de las longitudes de los enlaces.
    connection_matrix(topology)
        Construye la matriz de adyacencia.
    """

    @staticmethod
    def node_degree(topology: "Topology") -> Dict[int, int]:
        """
        Calculates the degree of each node in the topology.

        :param topology: Instance of topolib.topology.topology.Topology.
        :type topology: topolib.topology.topology.Topology
        :return: Dictionary {node_id: degree}
        :rtype: dict[int, int]
        """
        degree = {n.id: 0 for n in topology.nodes}
        for link in topology.links:
            degree[link.source.id] += 1
            degree[link.target.id] += 1
        return degree

    @staticmethod
    def link_length_stats(topology: "Topology") -> Dict[str, Optional[float]]:
        """
        Calculates the minimum, maximum, and average link lengths.

        :param topology: Instance of topolib.topology.topology.Topology.
        :type topology: topolib.topology.topology.Topology
        :return: Dictionary with keys 'min', 'max', 'avg'.
        :rtype: dict[str, float | None]
        """
        lengths = [l.length for l in topology.links]
        if not lengths:
            return {"min": None, "max": None, "avg": None}
        return {
            "min": min(lengths),
            "max": max(lengths),
            "avg": sum(lengths) / len(lengths),
        }

    @staticmethod
    def connection_matrix(topology: "Topology") -> List[List[int]]:
        """
        Builds the adjacency matrix of the topology.

        :param topology: Instance of topolib.topology.topology.Topology.
        :type topology: topolib.topology.topology.Topology
        :return: Adjacency matrix (1 if connected, 0 otherwise).
        :rtype: list[list[int]]
        """
        id_to_idx = {n.id: i for i, n in enumerate(topology.nodes)}
        size = len(topology.nodes)
        matrix = [[0] * size for _ in range(size)]
        for link in topology.links:
            i = id_to_idx[link.source.id]
            j = id_to_idx[link.target.id]
            matrix[i][j] = 1
            matrix[j][i] = 1
        return matrix



import pytest
from topolib.elements.node import Node
from topolib.elements.link import Link


def test_link_creation_and_properties():
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    link = Link(10, n1, n2, 5.5)

    assert link.id == 10
    assert link.source is n1
    assert link.target is n2
    assert link.length == 5.5


def test_link_length_validation():
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    with pytest.raises(ValueError):
        Link(1, n1, n2, -1)
    with pytest.raises(TypeError):
        Link(1, n1, n2, "not-a-number")


def test_link_endpoint_validation():
    n1 = Node(1, "A", 0.0, 0.0)

    # a dummy object without node attributes
    class Dummy:
        pass

    with pytest.raises(TypeError):
        Link(1, Dummy(), n1, 1.0)
    with pytest.raises(TypeError):
        Link(1, n1, Dummy(), 1.0)


def test_endpoints_and_repr():
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    link = Link(2, n1, n2, 10)
    assert link.endpoints() == (n1, n2)
    assert "Link(id=2" in repr(link)


def test_link_setters_and_errors():
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    link = Link(99, n1, n2, 1.0)
    # Test id setter
    link.id = 123
    assert link.id == 123

    # Test source setter error
    class Dummy:
        pass

    with pytest.raises(TypeError):
        link.source = Dummy()
    # Test target setter error
    with pytest.raises(TypeError):
        link.target = Dummy()
    # Test length setter error (non-numeric)
    with pytest.raises(TypeError):
        link.length = "bad"
    # Test length setter error (negative)
    with pytest.raises(ValueError):
        link.length = -5


def test_link_length_setter_typeerror():
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    link = Link(1, n1, n2, 1.0)
    with pytest.raises(TypeError):
        link.length = object()  # No convertible a float, cubre el except



import pytest
from topolib.topology.path import Path


class DummyNode:
    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return f"Node({self.id})"


class DummyLink:
    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return f"Link({self.id})"


def test_path_creation():
    nodes = [DummyNode(1), DummyNode(2), DummyNode(3)]
    links = [DummyLink("a"), DummyLink("b")]
    path = Path(nodes, links)
    assert path.nodes == nodes
    assert path.links == links
    assert path.length() == 2
    assert path.hop_count() == 2
    assert path.endpoints() == (nodes[0], nodes[-1])


def test_path_invalid():
    with pytest.raises(ValueError):
        Path([], [])
    with pytest.raises(ValueError):
        Path([DummyNode(1)], [])
    with pytest.raises(ValueError):
        Path([DummyNode(1), DummyNode(2)], [DummyLink("a"), DummyLink("b")])


def test_path_repr():
    nodes = [DummyNode(1), DummyNode(2)]
    links = [DummyLink("a")]
    path = Path(nodes, links)
    s = repr(path)
    assert s.startswith("Path(nodes=")



"""
Node class for optical network topologies.

This module defines the Node class, representing a network node with geographic coordinates.
"""

from typing import Tuple


class Node:
    """
    .. :noindex:

    Represents a node in an optical network topology.

    :param id: Unique identifier for the node.
    :type id: int
    :param name: Name of the node.
    :type name: str
    :param latitude: Latitude coordinate of the node.
    :type latitude: float
    :param longitude: Longitude coordinate of the node.
    :type longitude: float
    """

    def __init__(self, id: int, name: str, latitude: float, longitude: float):
        self._id = id
        self._name = name
        self._latitude = latitude
        self._longitude = longitude

    @property
    def id(self) -> int:
        """
        Get the unique identifier of the node.

        :return: Node ID.
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, value: int) -> None:
        """
        Set the unique identifier of the node.

        :param value: Node ID.
        :type value: int
        """
        self._id = value

    @property
    def name(self) -> str:
        """
        Get the name of the node.

        :return: Node name.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """
        Set the name of the node.

        :param value: Node name.
        :type value: str
        """
        self._name = value

    @property
    def latitude(self) -> float:
        """
        Get the latitude coordinate of the node.

        :return: Latitude value.
        :rtype: float
        """
        return self._latitude

    @latitude.setter
    def latitude(self, value: float) -> None:
        """
        Set the latitude coordinate of the node.

        :param value: Latitude value.
        :type value: float
        """
        self._latitude = value

    @property
    def longitude(self) -> float:
        """
        Get the longitude coordinate of the node.

        :return: Longitude value.
        :rtype: float
        """
        return self._longitude

    @longitude.setter
    def longitude(self, value: float) -> None:
        """
        Set the longitude coordinate of the node.

        :param value: Longitude value.
        :type value: float
        """
        self._longitude = value

    def coordinates(self) -> Tuple[float, float]:
        """
        Returns the (latitude, longitude) coordinates of the node.

        :return: Tuple containing latitude and longitude.
        :rtype: Tuple[float, float]
        """
        return self._latitude, self._longitude



import pytest
from topolib.elements.node import Node


def test_node_creation():
    node = Node(1, "A", -33.45, -70.66)
    assert node.id == 1
    assert node.name == "A"
    assert node.latitude == -33.45
    assert node.longitude == -70.66


def test_node_setters():
    node = Node(0, "", 0.0, 0.0)
    node.id = 10
    node.name = "TestNode"
    node.latitude = 55.5
    node.longitude = -120.1
    assert node.id == 10
    assert node.name == "TestNode"
    assert node.latitude == 55.5
    assert node.longitude == -120.1


def test_node_coordinates():
    node = Node(2, "B", 10.0, 20.0)
    coords = node.coordinates()
    assert coords == (10.0, 20.0)



"""
Unit tests for the Topology class.
Covers all public methods, success and error cases, and adjacency matrix validation (numpy), including NetworkX integration.
"""

import pytest
import numpy as np
from topolib.topology import Topology
from topolib.elements.node import Node
from topolib.elements.link import Link


def test_add_and_get_nodes():
    topo = Topology()
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    topo.add_node(n1)
    topo.add_node(n2)
    assert n1 in topo.nodes and n2 in topo.nodes
    assert len(topo.nodes) == 2
    # Check that nodes are present in the internal networkx graph
    assert n1.id in topo._graph.nodes  # type: ignore
    assert n2.id in topo._graph.nodes  # type: ignore


def test_add_duplicate_node_allowed():
    # The current design allows duplicate nodes (no check in Topology)
    topo = Topology()
    n1 = Node(1, "A", 0.0, 0.0)
    topo.add_node(n1)
    topo.add_node(n1)
    assert topo.nodes.count(n1) == 2


def test_remove_node():
    topo = Topology()
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    topo.add_node(n1)
    topo.add_node(n2)
    topo.remove_node(n1.id)
    assert n1 not in topo.nodes
    assert n2 in topo.nodes
    # Check that the node was removed from the internal graph
    assert n1.id not in topo._graph.nodes  # type: ignore


def test_remove_nonexistent_node():
    import networkx as nx

    topo = Topology()
    n1 = Node(1, "A", 0.0, 0.0)
    # Should raise NetworkXError when trying to remove a node not in the graph
    with pytest.raises(nx.NetworkXError):
        topo.remove_node(n1.id)
    assert n1 not in topo.nodes


def test_add_and_get_links():
    topo = Topology()
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    topo.add_node(n1)
    topo.add_node(n2)
    l1 = Link(1, n1, n2, 10.0)
    topo.add_link(l1)
    assert l1 in topo.links
    assert len(topo.links) == 1
    # Check that the edge is present in the internal networkx graph
    assert topo._graph.has_edge(n1.id, n2.id)  # type: ignore


def test_add_link_with_missing_nodes():
    # The current design allows adding links even if nodes are not in the topology
    topo = Topology()
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    l1 = Link(1, n1, n2, 10.0)
    topo.add_link(l1)
    assert l1 in topo.links


def test_remove_link():
    topo = Topology()
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    topo.add_node(n1)
    topo.add_node(n2)
    l1 = Link(1, n1, n2, 10.0)
    topo.add_link(l1)
    topo.remove_link(l1.id)
    assert l1 not in topo.links
    # Check that the edge was removed from the internal graph
    assert not topo._graph.has_edge(n1.id, n2.id)  # type: ignore


def test_remove_nonexistent_link():
    topo = Topology()
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    topo.add_node(n1)
    topo.add_node(n2)
    # Should not raise, just does nothing
    topo.remove_link(999)
    assert len(topo.links) == 0


def test_adjacency_matrix():
    topo = Topology()
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    n3 = Node(3, "C", 2.0, 2.0)
    topo.add_node(n1)
    topo.add_node(n2)
    topo.add_node(n3)
    l1 = Link(1, n1, n2, 10.0)
    l2 = Link(2, n2, n3, 20.0)
    topo.add_link(l1)
    topo.add_link(l2)
    matrix = topo.adjacency_matrix()
    assert isinstance(matrix, np.ndarray)
    assert matrix.shape == (3, 3)
    # Check connections
    idx = {n.id: i for i, n in enumerate(topo.nodes)}
    assert matrix[idx[n1.id], idx[n2.id]] == 1
    assert matrix[idx[n2.id], idx[n3.id]] == 1
    assert matrix[idx[n1.id], idx[n3.id]] == 0
    # Validate with networkx adjacency
    adj_nx = np.asarray(
        np.array(
            [
                [1 if topo._graph.has_edge(n1.id, n2.id) else 0 for n2 in topo.nodes]  # type: ignore
                for n1 in topo.nodes
            ]
        )
    )
    assert np.array_equal(matrix, adj_nx)


def test_adjacency_matrix_empty():
    topo = Topology()
    matrix = topo.adjacency_matrix()
    assert isinstance(matrix, np.ndarray)
    assert matrix.shape == (0, 0)



# Configuration file for the Sphinx documentation builder.

# -- Add project root to sys.path for autodoc --
import os
import sys
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '../../')))
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'topolib'
copyright = '2025, Danilo Borquez'
author = 'Danilo Borquez'
release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
]

templates_path = ['_templates']
exclude_patterns = []

language = 'en'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



import pytest

from topolib.analysis.metrics import Metrics
from topolib.topology import Topology


from topolib.elements.node import Node
from topolib.elements.link import Link


def test_node_degree():
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    n3 = Node(3, "C", 2.0, 2.0)
    l1 = Link(1, n1, n2, 10)
    l2 = Link(2, n2, n3, 20)
    topo = Topology(nodes=[n1, n2, n3], links=[l1, l2])
    result = Metrics.node_degree(topo)
    assert result == {1: 1, 2: 2, 3: 1}


def test_link_length_stats():
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    l1 = Link(1, n1, n2, 10)
    l2 = Link(2, n2, n1, 20)
    topo = Topology(nodes=[n1, n2], links=[l1, l2])
    stats = Metrics.link_length_stats(topo)
    assert stats["min"] == 10
    assert stats["max"] == 20
    assert stats["avg"] == 15
    # Empty case
    topo_empty = Topology()
    assert Metrics.link_length_stats(topo_empty) == {
        "min": None,
        "max": None,
        "avg": None,
    }


def test_connection_matrix():
    n1 = Node(1, "A", 0.0, 0.0)
    n2 = Node(2, "B", 1.0, 1.0)
    n3 = Node(3, "C", 2.0, 2.0)
    l1 = Link(1, n1, n2, 10)
    l2 = Link(2, n2, n3, 20)
    topo = Topology(nodes=[n1, n2, n3], links=[l1, l2])
    matrix = Metrics.connection_matrix(topo)
    assert matrix == [
        [0, 1, 0],
        [1, 0, 1],
        [0, 1, 0],
    ]



import importlib

def test_package_exports_version():
    pkg = importlib.import_module("topolib")
    assert hasattr(pkg, "__version__"), "topolib should expose __version__"
    assert isinstance(pkg.__version__, str), "__version__ should be a string"
